import java.lang.*;
import java.awt.*;

public class Cell {  

	//each cell's value is defined by a type
	//the four different types are as follows
	public static final int REAL = 0;
	public static final int INTEGER = 1;
  public static final int LABEL = 2;
  public static final int FORMULA = 3;
    
  boolean	needRedisplay;										//cell needs to be redisplayed
	boolean selected;                         //cell is selected, highlight cell
  int	type = INTEGER;												//cell type as defined by above
  String cellString = "";										//string eqivalent of cell value
  Color	bgColor;														//cell background color
  Color	fgColor;														//cell foreground color
  Color	highlightColor;											//cell highlight color
  int		width;															//cell's width in pixels
  int		height;															//cell's height in pixels
	double numeric = 0;
  SpreadSheetCanvas ssc;										//spreadsheet that contains the cell

	FormulaParser fp;

	public Cell(SpreadSheetCanvas ssc,	
		Color bgColor, Color fgColor,
		Color highlightColor,
		int width, int height) {

		this.ssc = ssc;
		this.bgColor = bgColor;
		this.fgColor = fgColor;
		this.highlightColor = highlightColor;
		this.width = width;
		this.height = height;
		selected = false;
		needRedisplay = true;    
		fp = new FormulaParser(ssc);
	}

	public void setUpdate(boolean b){
		needRedisplay = true;
		}

	public boolean isSelected(){
		return selected;
		}

	private boolean isInt(){
		try{
			Integer.parseInt(cellString);
			return true;
		}catch (Exception e){
			return false;
		}
	}

	private boolean isReal(){
		try{
			Float.valueOf(cellString);
			return true;
		}catch (Exception e){
			return false;
		}
	}

	private boolean isFormula(){
		try{
			char c = cellString.charAt(0);
			return (c == '=');
		}catch (Exception e){
			return false;
		}
	}

	public void recalculate(){
		try{
		if(type == FORMULA){
			numeric = fp.compute(cellString.substring(1));
			needRedisplay = true;
		}
		}catch (Exception e){}
	}

	public void setString(String str){
		cellString = str;
		if(isInt()){
			type = INTEGER;
			numeric = Double.valueOf(cellString).doubleValue();
		}else if(isReal()){
			type = REAL;
			numeric = Double.valueOf(cellString).doubleValue();
		}else if(isFormula()){
			type = FORMULA;
			cellString = cellString.toUpperCase();
			numeric = fp.compute(cellString.substring(1));
			//System.out.println("computed: " + numeric + " " + cellString.substring(1));
		}else{
			type = LABEL;
			numeric = 0;
		}
		needRedisplay = true;
	}

	public double getNumericValue(){
		return numeric;
	}

	public void setWidth(int nWidth){
		if (nWidth > 0) {
			this.width = nWidth;
		}
	}

	public int getWidth(){			
		return width;
	}

	public String getString() {
		return cellString;  
	}  

	public int getType(){
		return type;
	}
    
	public void select() {	
		selected = true;
		needRedisplay = true;
		//ssc.repaint();
	}

	public void deselect() {
		selected = false;
		needRedisplay = true;
		//ssc.repaint();   
	}
    

	public void paint(Graphics g, int x, int y) {	
		if(needRedisplay){
			FontMetrics fm = g.getFontMetrics();
			String tmpStr = cellString;
			if (selected) {	    
				g.setColor(highlightColor);	
				g.fillRect(x+2, y+3, width-2, height-3);
			}else {	    
				g.setColor(bgColor);	
				g.fillRect(x+2, y+3, width-2, height-3);
			}	
			if ((tmpStr != null) && needRedisplay){
		    switch (type) {
				  case Cell.REAL:
		      case Cell.INTEGER:
						g.setColor(Color.blue);
						break;
		      case Cell.LABEL:		
						g.setColor(fgColor);		
						break;
		      case Cell.FORMULA:		
						g.setColor(Color.red.darker());		
						Double tmp = new Double(numeric);
						int intval = tmp.intValue();
						//System.out.println("Int Val: " + intval);					
						if(intval == numeric){
							tmpStr = Integer.toString(intval);
						}else{
							tmpStr = Double.toString(numeric);
						}					
						break;	        
				}	    
				if (fm.stringWidth(tmpStr)+4 > width) {
					float char_per_pix = (float)tmpStr.length()/fm.stringWidth(tmpStr);
					int num_char = (new Float(width * char_per_pix)).intValue() - 1;
					if(num_char >0){
						g.drawString(tmpStr.substring(0, num_char),				 
							x+4, y + (height / 2) + 6);	
					}
				} else {		    				
					g.drawString(tmpStr, x+4, y + (height / 2) + 6);		
				}	
			}	
		}
		needRedisplay = false;    
	}
}
