//import java.lang.*;
import java.util.*;

/*
 * FormulaParser
 */
public class FormulaParser{
	
	public static final int DOUBLE = 0;
	public static final int FUNCTION = 1;
	public static final int DELIMITER = 2;
	public static final int CELL = 3;
	public static final int UNKNOWN = 4;

	String expression = null;
	StringTokenizer parser = null;
	Function func = null;
	String token = null;
	int type = UNKNOWN;

	SpreadSheetCanvas ssc;

	public FormulaParser(SpreadSheetCanvas s){		
		ssc = s;
	}

	private boolean isNum(String s){
		try{
			Double.valueOf(s);
		}catch (Exception e){
			return false;
		}
		return true;		
	}

	private boolean isDelim(String s){
		try{
			char tmp = s.charAt(0);
			if((tmp == '+') || (tmp == '-') || (tmp == '/') || (tmp == '*') ||
				(tmp == '^') || (tmp == '(') || (tmp == ')')){
				return true;
			}	
		}catch (NullPointerException Exception){
			return false;
		}
		return false;
	}

	private boolean isFunc(String s){
		try{
			if(s.charAt(0) == '@'){
				return true;
			}else{
				return false;
			}
		}catch (NullPointerException Exception){
			return false;
		}
	}


	private boolean isCell(String s){
		try{
			char c = s.charAt(0);
			int rest = Integer.parseInt(s.substring(1));
			if((c >= 'A') && (c <= 'Z') && (rest >= 0) && (rest <= Integer.MAX_VALUE)){
				return true;
			}else{
				return false;
			}
		}catch (Exception e){
			return false;
		}
	}

	private boolean isRange(String s){
		try{
			StringTokenizer st = new StringTokenizer(s, ":");
			String leftCell = st.nextToken();
			String rightCell = st.nextToken();
			//System.out.println("isRange: " + s + " left: " + leftCell + " right: " + rightCell);
			if((isCell(leftCell)) && (isCell(rightCell))){
				return true;
			}else{
				return false;
			}
		}catch (Exception e){
			return false;
		}
	}


	private void getToken(){
		try{
			token = parser.nextToken();
			//some problems were encountered with the StrinkTokenizer returning a " " 
			//even though a " " is not one of the delimiters.
			if(token.equals(" ")){
				token = parser.nextToken();
			}
			token = (new String(token)).trim();
			//System.out.println("Token: \"" + token + "\"");
			type = UNKNOWN;
			if(isNum(token)){
				type = DOUBLE;
			}else if(isDelim(token)){
				type = DELIMITER;
			}else if(isFunc(token)){
				type = FUNCTION;
			}else if(isCell(token)){
				type = CELL;
			}
		}catch(NoSuchElementException Exception){
			token = null;
			return;
		}
	}

	private double addSub(){
		double value = multDiv();
		try{
			char op = token.charAt(0);
			while((op == '+') || (op == '-')){
				getToken();
				double val = multDiv();
				if(op == '+'){
					value += val;
				}else{
					value -= val;
				}
				op = token.charAt(0);
			}
			return value;
		}catch (Exception e){
			//e.printStackTrace();
			return value;
		}
	}

	private double multDiv(){
		double value = pow();
		try{			
			char op = token.charAt(0);
	
			while((op == '*') || (op == '/')){
				getToken();
				double val = pow();
				if(op == '*'){
					value *= val;
				}else{
					value /= val;
				}
				op = token.charAt(0);
			}
			return value;
		}catch (Exception e){
			//e.printStackTrace();
			return value;
		}
	}

	private double pow(){
		double value = unary();
		try{			
			if(token.charAt(0) == '^'){
				getToken();
				double val = pow();
				value = new Double(Math.pow(value, val)).doubleValue();
			}
			return value;
		}catch(Exception e){
			//e.printStackTrace();
			return value;
		}
	}

	private double unary(){
		char op = '0';
		if(type == DELIMITER){
			op = token.charAt(0);
			if((op == '+') || (op == '-')){
				getToken();
			}
		}
		double value = paren();
		if(op == '-'){
			return (-value);
		}else{
			return (value);
		}
	}

	private double paren(){
		double value=0;
		if((type == DELIMITER) && (token.charAt(0) == '(')){
			getToken();
			value = addSub();
			if(token.charAt(0) != ')'){
				System.out.println("No matching parentheses.");
			}
		}else{
			switch(type){
			case DOUBLE:
				value = Double.valueOf(token).doubleValue();
				//System.out.println(value);
				break;
			case CELL:
				int row = Integer.parseInt(token.substring(1)) - 1;
				int column = token.charAt(0) - 'A';
				value = ssc.getCellValue(row, column);
				//System.out.println("row: " + row + " column: " + column);
				//System.out.println("cell: " + token + " = " + value);
				break;
			case FUNCTION:
				getToken();
				String name = token;
				getToken();
				getToken();
				double val;
				if(isRange(token)){
					value = func.evaluate(name, token);
					getToken();
					
					if(token.charAt(0) != ')'){
						System.out.println("No matching parentheses.");
					}

				}else{
					//value = paren();
					value = addSub();					
					
					if(token.charAt(0) != ')'){
						System.out.println("No matching parentheses.");
					}
					
					value = func.evaluate(name, value);
				}
				//System.out.println("value: " + value);
				//return value;
				break;
			default:
				break;
			}
		}
		getToken();
		return value;
	}


	public double compute(String str){	
		try {
			func = new Function(ssc);
			expression = str;
			parser = new StringTokenizer(expression, "+-*/^()@", true);
			getToken();
		}catch (Exception e){
			//e.printStackTrace();
		}		
		return addSub();
	}
}

