import java.awt.*;
import java.util.*;

/*
 *
 * SpreadSheetCanvas
 *
 */
public class SpreadSheetCanvas extends Canvas{
	final int   m_nPrecision = 2;      //# pixels the cursor has to be near
	                                   //the drag line in order to drag it
	final int   m_nReset = -1;         //number used to test if a variable is
	                                   //resetted

	Font				cellFont;							 //font of cells
	Font				labelFont;						 //font of row and column labels
	
  Color		    cellColor;             //highlight cell color

  public int  cellWidth;             //default cell width
  public int  cellHeight;            //default cell height
  public int  rowLabelWidth;         //default row label width
  int			    rows;                  //total number of rows
  int			    columns;               //total number of columns
  int			    selectedRow;           //current selected row
  int			    selectedColumn;        //current selected column
	int					m_dragLineList[];      //stores array of x value of vertical lines
	int	        m_dragCell;            //cell being dragged
 
  boolean		  fullUpdate;            //update spreadsheet?
	boolean     m_draggingLine;				 //line being dragged?
    
  Vector	    cells;								 //vector of cells

	int oldx;													 //old value of the x axis generated by
	                                   //the mouse down and mouse drag methods

  InputField	inputArea;	           //input area of spreadsheet

	Vector			selectVec;						 //vector of points representing selected cells
																	   //x = columns
																		 //y = rows
	Vector			copiedVec;						 //a copy of selectVec used for copying and pasting
	Vector      copiedVecStr;					 //vector of strings corresponding to copiedVec
	Dimension   selectDim;
	Dimension   copiedDim;
	SpreadSheet ss;

	private Image m_osi;							 //off screen image
	private Graphics m_osg;						 //off screen graphics

	public SpreadSheetCanvas(int r, int c, SpreadSheet ss){
		super();
		rows = r;
		columns = c;
		this.ss = ss;
		init();
	}

	public void setInputField(InputField inf){
		inputArea = inf;
		}

  /*
  ** Initialize canvas
  */
  public synchronized void init() {
		cellColor = Color.yellow.brighter();

		cellFont = new Font("Courier", Font.BOLD, 12);
		labelFont = new Font("Courier", Font.PLAIN, 10);

		cellWidth = 100;
    cellHeight = 18;
    rowLabelWidth = 15;

		selectedRow = m_nReset;
		selectedColumn = m_nReset;
		m_dragCell = m_nReset;

		fullUpdate = false;
		m_draggingLine = false;

		m_dragLineList = new int[columns];

		setBackground(Color.lightGray);
		cells = new Vector();

		selectVec = new Vector();
		copiedVec = new Vector();
		copiedVecStr = new Vector();
		selectDim = new Dimension();
		copiedDim = new Dimension();

		Cell c;
		//initialize appropriate cells
		for (int i=0; i < rows; i++) {
			for (int j=0; j < columns; j++) { 
				c =  new Cell(this,			
					Color.lightGray,
					Color.black,
					cellColor,
					cellWidth,
					cellHeight);						
				cells.addElement(c);	
			}		 
		}	

		m_dragLineList = new int[columns];
		m_dragLineList[0] = rowLabelWidth + ((Cell)cells.elementAt(0)).getWidth();
		for (int n=1; n < columns; n++){                    					
			m_dragLineList[n] = m_dragLineList[n-1] + cellWidth;
		}

	}
 

	public void recalculate(){
		for (int i=0; i < rows; i++) {
			for (int j=0; j < columns; j++) {
				((Cell)cells.elementAt(j % columns + i*columns)).recalculate();				
			}
		}
	}

	/*
	 * Sets the width for an entire column
	 */
	public void setColumnWidth(int nColumn, int nWidth){
		for(int i = 0; i < rows; i++){
			((Cell)cells.elementAt(nColumn % columns + i*columns)).setWidth(nWidth);			
		}
	}

	/*
	 * Returns the width of a certain column
	 */
	public int getColumnWidth(int nColumn){
		if ((rows>=0) && (columns >= nColumn)){			
			return ((Cell)cells.elementAt(nColumn % columns + 0*columns)).getWidth();
		} else {
			return m_nReset; //error condition
		}
	}

	public void setCellValue(int cellRow, int cellColumn, String valu){
		int position = cellColumn % columns + cellRow * columns;
		if ( position <= cells.size()){
			((Cell)cells.elementAt(position)).setString(valu);
			repaint(); 	 
		}
	}

	public void setCellValue(int cellRow, int cellColumn, double valu){
		int position = cellColumn % columns + cellRow * columns;
		if ( position <= cells.size()){
			String val = new String(String.valueOf(valu));
			((Cell)cells.elementAt(position)).setString(val.toString());
			repaint(); 	 
		}
	}


	/*
	 * Set current cell value
	 */
	 
	public void setCellValue(String s) {
		if (selectedRow == m_nReset || selectedColumn == m_nReset) {				
			return;		 
		}	
		((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).setString(s);		
		repaint(); 	 
	} 

	public double getCellValue(int r, int c){
		if ((r != m_nReset || c != m_nReset) &&
			(r < rows) && (c < columns)){		
			return ((Cell)cells.elementAt(c % columns + r*columns)).getNumericValue();			
		}else{
			return 0;
		}
	}
      

	public void selectUpperCell(){
		if((selectedRow != m_nReset) && (selectedColumn != m_nReset)){
			if(selectedRow == 0){
				selectedRow = rows-1;
			}else{
				selectedRow = selectedRow - 1;
			}
			deselectVectorCells();
			Point p = new Point(selectedColumn, selectedRow);
			selectVec.addElement(p);
			((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).select();			
			inputArea.setText( new String(((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).getString()));
		  inputArea.requestFocus();
		  repaint();	
		}
	}

	public void selectLowerCell(){
		if((selectedRow != m_nReset) && (selectedColumn != m_nReset)){
			selectedRow = (selectedRow + 1) % rows;
			deselectVectorCells();
			Point p = new Point(selectedColumn, selectedRow);
			selectVec.addElement(p);
			((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).select();			
			inputArea.setText( new String(((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).getString()));
		  inputArea.requestFocus();
		  repaint();	
		}
	}

	public void selectLeftCell(){
		if((selectedRow != m_nReset) && (selectedColumn != m_nReset)){
			if(selectedColumn == 0){
				selectedColumn = columns -1;
			}else{
				selectedColumn = selectedColumn - 1;
			}
			deselectVectorCells();
			Point p = new Point(selectedColumn, selectedRow);
			selectVec.addElement(p);
			((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).select();			
			inputArea.setText(new String( ((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).getString()));
		  inputArea.requestFocus();
		  repaint();	
		}
	}

	public void selectRightCell(){
		if((selectedRow != m_nReset) && (selectedColumn != m_nReset)){
			selectedColumn = (selectedColumn + 1) % columns;
			deselectVectorCells();
			Point p = new Point(selectedColumn, selectedRow);
			selectVec.addElement(p);
			((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).select();			
			inputArea.setText(new String(((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).getString()));
		  inputArea.requestFocus();
		  repaint();	
		}
	}


	/*
	 * Called before paint method
	 */
	public void update(Graphics g) { 
		if (! fullUpdate) {	
			m_osg.setFont(cellFont);
			g.setFont(cellFont);
			int cx = rowLabelWidth;
			int cy;
			for (int i=0; i < rows; i++) {
			  for (int j=0; j < columns; j++) {
					cy = ((i+1) * cellHeight);
					if (((Cell)cells.elementAt(j % columns + i*columns)) != null) {						
						((Cell)cells.elementAt(j % columns + i*columns)).paint(g, cx, cy);
					}
					if (j == columns - 1 ) {
						cx = rowLabelWidth;
					}else {										
						cx += ((Cell)cells.elementAt(j % columns + i*columns)).getWidth();
					}		   
				}
			}
		}else {
			Dimension d = getSize();                           
			m_osg.setColor(getBackground());
			m_osg.fillRect(0,0, d.width, d.height);
			paint(g);    			 
		  fullUpdate = false;			 
		}	
	}
    
	public synchronized void paint(Graphics g) {
		int i, j;
		int cx = 0;
		int cy = 0;
		char l[] = new char[1];
		Dimension d = getSize();

		if(( m_osi == null ) && (m_osg == null)){
			m_osi = createImage(getSize().width, getSize().height);
			m_osg = m_osi.getGraphics();
		}

		m_osg.setColor(Color.pink);
		m_osg.fillRect(0, 0, d.width, cellHeight);
		m_osg.fillRect(0, 0, rowLabelWidth, d.height);

		

		//paint cell value
		m_osg.setFont(cellFont);
		cx = rowLabelWidth;
		for (i=0; i < rows; i++) {
	    for (j=0; j < columns; j++) {
				cy = ((i+1) * cellHeight);
				((Cell)cells.elementAt(j % columns + i*columns)).setUpdate(true);				
				if (((Cell)cells.elementAt(j % columns + i*columns)) != null) {
					((Cell)cells.elementAt(j % columns + i*columns)).paint(m_osg, cx, cy);					
				}
				if (j == columns - 1 ) {
					cx = rowLabelWidth;
				}else {				
					cx += ((Cell)cells.elementAt(j % columns + i*columns)).getWidth();					
				}
	    }
		}

		//paint horizontal lines
		for (i=0; i < rows+1; i++) {	    		    
			m_osg.setColor(getBackground());
			cy = (i+1) * cellHeight;	            	            
			m_osg.draw3DRect(0, cy, d.width, 2, true);		    
			//paint row label            
			if (i < rows) {	
				m_osg.setFont(labelFont);
				m_osg.setColor(Color.blue.darker());		        	     
				m_osg.drawString("" + (i+1), 2, cy + 14);	
			}			
		}					
		
		m_osg.setColor(getBackground());

		//paint leftmost vertical line
		m_osg.draw3DRect(rowLabelWidth, 0, 1, d.height, true);

		//paint vertical lines	
		cx = 1;
		for (i=0; i < columns; i++) {	
			//paint column label
			if (i < columns) {
				m_osg.setFont(labelFont);
				m_osg.setColor(Color.blue.darker());		
				l[0] = (char)((int)'A' + i);		
				m_osg.drawString(new String(l),			     
					cx + rowLabelWidth + (((Cell)cells.elementAt(i % columns+ 0*columns)).getWidth() / 2), cellHeight*4/5);
			}
			m_osg.setColor(getBackground());		
			cx += ((Cell)cells.elementAt(i % columns+ 0*columns)).getWidth();			
			m_osg.draw3DRect(cx + rowLabelWidth-(m_nPrecision/2), 0, 1, d.height, true);	
		}		
		
		g.drawImage(m_osi, 0, 0, this);
	}

	public boolean mouseUp(Event evt, int x, int y){
		if ((m_draggingLine == true) && (m_dragCell != m_nReset)){
			int leftline = rowLabelWidth;
			for(int i=m_dragCell-1; i>=0; i--){				
				leftline += getColumnWidth(i);
			}
			int offset = x - leftline;			

			//cell can't have negative width
			if (offset > 0){
				//larger resize
				if(offset > getColumnWidth(m_dragCell)){
					int incSize = offset - getColumnWidth(m_dragCell);
					for (int i=m_dragCell+1; i<columns; i++){
						m_dragLineList[i] += incSize;
					}
					setColumnWidth(m_dragCell, offset);
					m_dragLineList[m_dragCell] = offset + leftline;
				}else if(offset < getColumnWidth(m_dragCell)){
					int decSize = getColumnWidth(m_dragCell) - offset;
					for (int i=m_dragCell+1; i<columns; i++){
						m_dragLineList[i] -= decSize;
					}
					setColumnWidth(m_dragCell, offset);
					m_dragLineList[m_dragCell] = offset + leftline;				
				}
			}
			m_draggingLine = false;
			m_dragCell = m_nReset;
			fullUpdate = true;
			repaint();
			return true;
		}
		return false;
	}

	private void drawDragBorder(int col){
	Graphics g = getGraphics();
	if ( g != null )  
		{
			g.setXORMode(Color.white);			
			g.drawLine(col, 0, col, getSize().height);
		}
	g.dispose();
	}

	public boolean mouseDrag( Event evt, int x, int y ){
		if ((m_draggingLine == true) && (m_dragCell != m_nReset)){
			drawDragBorder(oldx);
			drawDragBorder(x);
		}else{ //select multiple cells
			calculateSelectedRow(y);
			calculateSelectedColumn(x);
			if((selectedRow == m_nReset) || (selectedColumn == m_nReset)){
				deselectVectorCells();
				repaint();
			}else{
				Point p = new Point(selectedColumn, selectedRow);
				if(selectVec.size() > 0){					
					int baseRow = ((Point)selectVec.elementAt(0)).y;
					int baseCol = ((Point)selectVec.elementAt(0)).x;
					int limitRow = p.y;
					int limitCol = p.x;
					deselectVectorCells();
					repaint();
					//lower right select
					if((limitRow >= baseRow) && (limitCol >= baseCol)){
						for(int i = baseRow; i <= limitRow; i++){
							for(int j = baseCol; j <= limitCol; j++){
								p = new Point(j, i);
								selectVec.addElement(p);
								((Cell)cells.elementAt(j % columns + i*columns)).select();								
							}
						}
						selectDim.width = limitCol - baseCol + 1;
						selectDim.height = limitRow - baseRow + 1;
					//upper right select
					}else if((limitRow <= baseRow) && (limitCol >= baseCol)){
						for(int i = baseRow; i >= limitRow; i--){
							for(int j = baseCol; j <= limitCol; j++){
								p = new Point(j, i);
								selectVec.addElement(p);
								((Cell)cells.elementAt(j % columns + i*columns)).select();								
							}
						}
						selectDim.width = limitCol - baseCol + 1;
						selectDim.height = baseRow - limitRow + 1;
					//upper left select
					}else if((limitRow <= baseRow) && (limitCol <= baseCol)){
						for(int i = baseRow; i >= limitRow; i--){
							for(int j = baseCol; j >= limitCol; j--){
								p = new Point(j, i);
								selectVec.addElement(p);
								((Cell)cells.elementAt(j % columns + i*columns)).select();								
							}
						}
						selectDim.width = baseCol - limitCol + 1;
						selectDim.height = baseRow - limitRow+ 1;
					//lower left select
					}else if((limitRow >= baseRow) && (limitCol <= baseCol)){
						for(int i = baseRow; i <= limitRow; i++){
							for(int j = baseCol; j >= limitCol; j--){
								p = new Point(j, i);
								selectVec.addElement(p);
								((Cell)cells.elementAt(j % columns + i*columns)).select();								
							}
						}
						selectDim.width = baseCol - limitCol + 1;
						selectDim.height = limitRow - baseRow + 1;
					}
					repaint();
				}
			}
		}		
		oldx = x;
		return true;
	}


	public boolean mouseMove(Event evt, int x, int y){
		//int cursor = Frame.DEFAULT_CURSOR;
		Cursor cursor = new Cursor(Cursor.DEFAULT_CURSOR);
		if ((y > 0) &&
				(y < cellHeight)){
			for(int n=0; n<columns; n++){
				if (Math.abs(x-(m_dragLineList[n])) < m_nPrecision){
					cursor = Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR);
				}
			}
		}
		//getFrame().setCursor(cursor);
		this.setCursor(cursor);
		return true;
	}

	protected Frame getFrame(){
		Frame fr = null;

		if ( fr == null )
			  {
				Component c = this;
				while ( c != null && ! (c instanceof Frame) )
				c = c.getParent();
				fr = (Frame) c;
			  }
		return fr;
	}

	public boolean mouseDown(Event evt, int x, int y) {
		oldx = x;		

		calculateSelectedRow(y);
		calculateSelectedColumn(x);

		if ((y > 0) &&
				(y < cellHeight)){
				for(int n=0; n<columns; n++){
					if (Math.abs(x-(m_dragLineList[n])) < m_nPrecision){
						m_draggingLine = true;
						m_dragCell = n;						
						break;
					}				
				}		
		}else if((selectedRow == m_nReset) || (selectedColumn == m_nReset)){
			deselectVectorCells();
			repaint();
		}else{
			deselectVectorCells();
			Point p = new Point(selectedColumn, selectedRow);
			selectVec.addElement(p);
			((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).select();			
			selectDim.width = 1;
			selectDim.height = 1;
			inputArea.setText(new String(((Cell)cells.elementAt(selectedColumn % columns + selectedRow*columns)).getString()));
		  inputArea.requestFocus();
		  repaint();	
		}
		return true;    
	}

	private void deselectVectorCells(){
		int r;
		int c;	
		//System.out.println("vec size b4 removing: " + selectVec.size());
		if(selectVec.size() >0){
			for (int i = 0; i<selectVec.size(); i++){
				r = ((Point)selectVec.elementAt(i)).y;
				c = ((Point)selectVec.elementAt(i)).x;	
				//System.out.println(r+ " " + c);
				((Cell)cells.elementAt(c % columns + r * columns)).deselect();				
			}
			selectVec.removeAllElements();
			selectDim.width = 0;
			selectDim.height = 0;
		}
	}


	private void calculateSelectedRow(int y){
		selectedRow = ((y - cellHeight) / cellHeight);

		if ((selectedRow >= rows) || (y < cellHeight)) {	    
			selectedRow = m_nReset;
		}
	}

	private void calculateSelectedColumn(int x){
		int cx = rowLabelWidth + 2;

		if (x < rowLabelWidth) {
			selectedColumn = m_nReset;
		}else{
			for (int i = 0; i < columns; i++){
				if (cx > x) {					
					selectedColumn = i-1;
					break;
				}else {
					selectedColumn = m_nReset;
					cx += ((Cell)cells.elementAt(i % columns + 0 * columns)).getWidth();					
				}
				//evaluate last column
				if((i == (columns-1)) && (cx > x)){
					selectedColumn = i;
				}
			}
		}
	}

	public void handleCut(){
		handleCopy();
		handleClear();
		}

	public void handleCopy(){
		copiedVec = new Vector();
		copiedVecStr = new Vector();
		Point lc = findUpperLeftCorner();
		//System.out.println("ulc row: " + lc.y + " ulc col: " + lc.x);
		String s;
		for(int i=lc.y; i<selectDim.height + lc.y; i++){
			for(int j=lc.x; j<selectDim.width + lc.x; j++){				
				Point p = new Point(j, i);
				s = new String(((Cell)cells.elementAt(j % columns + i * columns)).getString());
				//System.out.println("i: " + i + " j: " + j);
				copiedVec.addElement(p);
				copiedVecStr.addElement(s);
				//System.out.println("'" + s + "'");
			}
		}
		copiedDim = new Dimension(selectDim.width, selectDim.height);
		}

		private Point findUpperLeftCorner(){
		Point ulc = new Point(0, 0); //upper left corner
		Point cp = new Point(0, 0);
		bye:{
			for(int i=0; i<rows; i++){
				for(int j=0; j< columns; j++){
					cp.move(j, i);
					if(selectVec.contains(cp)){
						ulc.move(j,i);
						break bye;
					}
				}
			}
		}//bye
		return ulc;
	}
		

	public void handlePaste(){
		if((copiedVecStr != null) && (selectVec != null) &&
			(!copiedVecStr.isEmpty()) && (!selectVec.isEmpty())){
			if(copiedVecStr.size() == 1){
				Point cv = (Point)copiedVec.elementAt(0);				
				String cvStr = (String)copiedVecStr.elementAt(0);
				//System.out.println("::" + cvStr + "'");
				if( ("".equals(cvStr)) || (cvStr.charAt(0) != '=') ){
					for(int n = 0; n < selectVec.size(); n++){
						Point p = (Point)selectVec.elementAt(n);
						((Cell)cells.elementAt(p.x % columns + p.y*columns)).setString(cvStr);
					}
				}else if(cvStr.charAt(0) == '='){ //handle pasting of formulas
					for(int n = 0; n < selectVec.size(); n++){
						Point p = (Point)selectVec.elementAt(n);						
						((Cell)cells.elementAt(p.x % columns + p.y*columns)).setString(pasteFormula(cvStr, cv, p));												
						//inputArea.setText("");						
						inputArea.requestFocus();
					}
				}
				Point lhc = findUpperLeftCorner();					
				inputArea.setText( new String(((Cell)cells.elementAt(lhc.x % columns + lhc.y * columns)).getString()));					
			}else if(selectVec.size() >= 1){
				Point sv = (Point)selectVec.elementAt(0);				
				if((selectVec.size() == 1) || (((selectDim.width % copiedDim.width) == 0) && 
					((selectDim.height % copiedDim.height) == 0))){
					String curStr;
					Point src = new Point(0, 0);
					Point des = new Point(0, 0);
					int i = 0;
					if(selectVec.size() == 1){
						deselectVectorCells();
						for(int r = sv.y; r < (sv.y+copiedDim.height); r++){
							for(int c = sv.x; c < (sv.x+copiedDim.width); c++){								
								curStr = new String((String)copiedVecStr.elementAt(i));
								//System.out.println("'" + curStr + "'");
								if( ("".equals(curStr)) || (curStr.charAt(0) != '=') ){								
									if((r < rows) && (c < columns)){
										((Cell)cells.elementAt(c % columns + r * columns)).setString(curStr);										
										selectVec.addElement(new Point(c, r));
										((Cell)cells.elementAt(c % columns + r*columns)).select();									
									}
								}else if(curStr.charAt(0) == '='){ //handle pasting of formulas
									if((r < rows) && (c < columns)){
										src.move(((Point)copiedVec.elementAt(i)).x, ((Point)copiedVec.elementAt(i)).y);
										des.move(c, r);
										((Cell)cells.elementAt(c % columns + r * columns)).setString(pasteFormula(curStr, src, des));										
										selectVec.addElement(new Point(c, r));
										((Cell)cells.elementAt(c % columns + r*columns)).select();
									}
								}
								i++;
							}
						}
						selectDim.width = copiedDim.width;
						selectDim.height = copiedDim.height;
					}else{
						int ini = 0;
						int ctn = 1;
						Point lhc = findUpperLeftCorner();
						for(int r = lhc.y; r < (lhc.y+selectDim.height); r++){
							for(int c = lhc.x; c < (lhc.x+selectDim.width); c++){								

								if((i != 0) && (( i % copiedDim.width) == 0)){
									i = ini;
								}
								
								curStr = new String((String)copiedVecStr.elementAt(i));
								//System.out.println("'" + curStr + "'" + i);
								if( ("".equals(curStr)) || (curStr.charAt(0) != '=') ){								
									if((r < rows) && (c < columns)){
										((Cell)cells.elementAt(c % columns + r * columns)).setString(curStr);										
										selectVec.addElement(new Point(c, r));
										((Cell)cells.elementAt(c % columns + r*columns)).select();										
									}
								}else if(curStr.charAt(0) == '='){
									if((r < rows) && (c < columns)){
										src.move(((Point)copiedVec.elementAt(i)).x, ((Point)copiedVec.elementAt(i)).y);
										des.move(c, r);
										((Cell)cells.elementAt(c % columns + r * columns)).setString(pasteFormula(curStr, src, des));										
										selectVec.addElement(new Point(c, r));
										((Cell)cells.elementAt(c % columns + r*columns)).select();									
									}
								}
								i++;
							}
							i = (ctn++) * copiedDim.width;
							if(i >= (copiedDim.width * copiedDim.height)){
								i = 0;
								ctn = 1;
							}
							ini = i;
						}
					}				
					Point lhc = findUpperLeftCorner();					
					inputArea.setText( new String(((Cell)cells.elementAt(lhc.x % columns + lhc.y * columns)).getString()));
				}else{
					Point lhc = findUpperLeftCorner();
					((Cell)cells.elementAt(lhc.x % columns + lhc.y * columns)).setString("! Dimensions not compatible!");					
				}				
			}
		}
		recalculate();
		repaint();
	}			

	private String pasteFormula(String formula, Point s, Point d){
		StringBuffer strbuf =  new StringBuffer();
		String token;
		StringTokenizer st = new StringTokenizer(formula, "+-*/^()@:=", true);		

		int ctn = st.countTokens();
		for(int i=0; i < ctn; i++){
			token = st.nextToken();			
			if(isCell(token)){
				int c = (int)token.charAt(0);
				int r = Integer.parseInt(token.substring(1));

				int rowOffset = s.y - r;		
				int newRow = d.y - rowOffset;
				if(newRow < 1){
					newRow = 1;
				}else if(newRow > rows){
					newRow = rows;
				}

				int colOffset = (s.x + ((int)'A')) - c;
				char newCol = (char)((d.x + ((int)'A')) - colOffset);
				if(newCol < 'A'){
					newCol = 'A';
				}else if(newCol > 'Z'){
					newCol = 'Z';
				}
				
				strbuf.append(newCol);
				strbuf.append(newRow);
			}else{
				strbuf.append(token);
			}
		}
		
		return strbuf.toString();
	}

	private boolean isCell(String s){
		try{
			char c = s.charAt(0);
			int rest = Integer.parseInt(s.substring(1));
			if((c >= 'A') && (c <= 'Z') && (rest >= 0) && (rest <= Integer.MAX_VALUE)){
				return true;
			}else{
				return false;
			}
		}catch (Exception e){
			return false;
		}
	}

	public void handleClear(){
		Point ulc = findUpperLeftCorner();
		for(int i = ulc.y; i < (ulc.y+selectDim.height); i++){
			for(int j = ulc.x; j < (ulc.x+selectDim.width); j++){
				((Cell)cells.elementAt(j % columns + i * columns)).setString("");
			}
		}	
		repaint();
		}

	private void selectColumns(int initCol, int endCol){
		Point p;
		for(int r=0; r < rows; r++){
			for(int c=initCol; c<=endCol; c++){
				((Cell)cells.elementAt(c % columns + r * columns)).select();
				p = new Point(c, r);
				selectVec.addElement(p);
			}
		}
		selectDim.width = columns - initCol;
		selectDim.height = rows;
	}

	private void selectRows(int initRow, int endRow){
		Point p;
		for(int r=initRow; r <= endRow; r++){
			for(int c=0; c < columns; c++){
				((Cell)cells.elementAt(c % columns + r * columns)).select();
				p = new Point(c, r);
				selectVec.addElement(p);
			}
		}
		selectDim.width = columns;
		selectDim.height = rows - initRow;
		//System.out.println("width: " + selectDim.width + " height: " + selectDim.height);
	}

	public void handleInsertRow(){
		int size;
		if(selectDim.width != 0){
			size = selectVec.size()/selectDim.width;
		}else{
			return;
		}
		//System.out.println("size: " + size);
		//System.out.println("bsv: " + selectVec.size());
		if((size > 0) && ((size + rows) <= 100)){
			Point ulc = findUpperLeftCorner();

			//System.out.println("ulc: " + ulc.x + " " + ulc.y);

			Vector backupSVec = new Vector();
			Vector backupCVec = new Vector();
			Vector backupCVecStr = new Vector();
			Dimension backupselectDim = new Dimension(selectDim.width, selectDim.height);
			Dimension backupcopiedDim = new Dimension(copiedDim.width, copiedDim.height);

			for (int i = 0; i < selectVec.size(); i++){
				backupSVec.addElement(selectVec.elementAt(i));
				//System.out.println("sv: " + selectVec.elementAt(i));
				//System.out.println("bv: " + backupSVec.elementAt(i));				
			}

			for (int n = 0; n < copiedVec.size(); n++){
				backupCVec.addElement(copiedVec.elementAt(n));
				backupCVecStr.addElement(copiedVecStr.elementAt(n));				
			}

			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			copiedDim.width = 0;
			copiedDim.height = 0;

			deselectVectorCells();
		
			//for(int t=0; t<cells.size(); t++){
			//	System.out.print( ((Cell)cells.elementAt(t)).getString() + " ");
			//}

			//System.out.println(" ");

			//int tmp = rows-1;
			//System.out.println("ulc.y: " + ulc.y + " r-1: " + tmp);
			selectRows(ulc.y, rows-1);
			
			handleCut();
	
			deselectVectorCells();
			
			rows += size;
			//setSize((cellWidth*columns) + rowLabelWidth, (cellHeight)*(rows));

			//index of row containing the upper left corner of the selection
			//int indexOfULC = (ulc.y + 1) * columns;
			//System.out.println(indexOfULC);

			int[] widthList = new int[columns];
			for(int t=0; t<columns; t++){
				widthList[t] = ((Cell)cells.elementAt(t)).getWidth();
			}

			//System.out.println(" ");
			for(int s=0; s < size; s++){
				for(int k=0; k<columns; k++){
					//System.out.print("k: " + k + " column width: " + widthList[k] + " ");
					Cell c = new Cell(this,			
						Color.lightGray,
						Color.black,
						cellColor,
						widthList[k],
						cellHeight);				
				
					//int where = indexOfULC - columns + k + s*columns;
					int where = ulc.y * columns + k + s*columns;
					//System.out.println("where: " + where);
					cells.insertElementAt(c, where);	
				}
			}

			Point pt = new Point(0, ulc.y + size);
			selectVec.addElement(pt);
			selectDim.width = 1;
			selectDim.height = 1;

			handlePaste();

			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			deselectVectorCells();
			
			for (int i = 0; i < backupSVec.size(); i++){				
				int ix = ((Point)backupSVec.elementAt(i)).x;
				int iy = ((Point)backupSVec.elementAt(i)).y + size;
				Point np = new Point(ix, iy);
				selectVec.addElement(np);
				int r = ((Point)selectVec.elementAt(i)).y;
				int c = ((Point)selectVec.elementAt(i)).x;
				((Cell)cells.elementAt(c % columns + r * columns)).select();
			}
			
			ulc = findUpperLeftCorner();
			inputArea.setText( new String(((Cell)cells.elementAt(ulc.x % columns + ulc.y * columns)).getString()));
			
			for (int n = 0; n < backupCVec.size(); n++){
				copiedVec.addElement(backupCVec.elementAt(n));
				copiedVecStr.addElement(backupCVecStr.elementAt(n));				
			}

			selectDim.width = backupselectDim.width;
			selectDim.height = backupselectDim.height;
			copiedDim.width = backupcopiedDim.width;
			copiedDim.height = backupcopiedDim.height;

			fullUpdate = true;
			repaint();
		}
	}

	public void handleInsertColumn(){
		int size;
		if(selectDim.height != 0){
			size = selectVec.size()/selectDim.height;
		}else{
			return;
		}
		//System.out.println("size: " + size);
		//System.out.println("bsv: " + selectVec.size());
		if((size > 0) && ((size + columns) <= 26)){
			Point ulc = findUpperLeftCorner();

			//System.out.println("ulc: " + ulc.y + " " + ulc.x);

			Vector backupSVec = new Vector();
			Vector backupCVec = new Vector();
			Vector backupCVecStr = new Vector();
			Dimension backupselectDim = new Dimension(selectDim.width, selectDim.height);
			Dimension backupcopiedDim = new Dimension(copiedDim.width, copiedDim.height);

			for (int i = 0; i < selectVec.size(); i++){
				backupSVec.addElement(selectVec.elementAt(i));
				//System.out.println("sv: " + selectVec.elementAt(i));
				//System.out.println("bv: " + backupSVec.elementAt(i));				
			}

			for (int n = 0; n < copiedVec.size(); n++){
				backupCVec.addElement(copiedVec.elementAt(n));
				backupCVecStr.addElement(copiedVecStr.elementAt(n));				
			}

			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			copiedDim.width = 0;
			copiedDim.height = 0;

			deselectVectorCells();
		
			//int tmp = rows-1;
			//System.out.println("ulc.y: " + ulc.y + " r-1: " + tmp);
			selectColumns(ulc.x, columns-1);
			
			handleCut();
	
			deselectVectorCells();
			
			//setSize((cellWidth*(columns + size)) + rowLabelWidth, (cellHeight)*(rows));

			//index of column containing the upper left corner of the selection
			int indexOfULC = ulc.x;
			//System.out.println("ulc.x: " + ulc.x);
			
			int where = 0;
			for(int s=0; s < size; s++){
				for(int k=0; k<rows; k++){
					Cell c = new Cell(this,			
						Color.lightGray,
						Color.black,
						cellColor,
						cellWidth,
						cellHeight);				
				
					//int where = indexOfULC + k*(columns + 1) + s;
					if(k == 0){
						int start = indexOfULC + s;
						where = start;
						//System.out.println("where: " + start);
						cells.insertElementAt(c, start);
					}else{
						where = where + (columns + 1) + s;
						cells.insertElementAt(c, where);
						//System.out.println("where: " + where);
					}
				}
			}

			columns += size;

			int tmpLineList[] = new int[columns];

			for(int n=0; n < tmpLineList.length; n++){
				//System.out.print(n);
				if(n < ulc.x){
					tmpLineList[n] = m_dragLineList[n];
					//System.out.println(" first");
				}else if(n >= (ulc.x + size)){
					//System.out.println(" second: "+n);
					if((ulc.x == 0) && (n == size)){
						tmpLineList[n] = tmpLineList[n-1] + m_dragLineList[n-size] - rowLabelWidth;
						setColumnWidth(n, m_dragLineList[n-size] - rowLabelWidth);
					}else{
						tmpLineList[n] = tmpLineList[n-1] + m_dragLineList[n-size] - m_dragLineList[n-size-1];
						setColumnWidth(n, m_dragLineList[n-size]- m_dragLineList[n-size-1]);
					}
				}else{
					//System.out.println(" third");
					if(n != 0){
						tmpLineList[n] = tmpLineList[n-1] + cellWidth;
					}else{
						tmpLineList[n] = cellWidth+rowLabelWidth;
					}
					setColumnWidth(n, cellWidth);
				}
			}
			
			//for(int t=0; t < tmpLineList.length; t++){
			//	System.out.print(" " + tmpLineList[t]);
			//}

			//System.out.println(" " );

			m_dragLineList = new int[columns];
			System.arraycopy(tmpLineList, 0, m_dragLineList, 0, tmpLineList.length); 
			
			//for(int t=0; t < m_dragLineList.length; t++){
			//	System.out.print(" " + m_dragLineList[t]);
			//}

			Point pt = new Point(ulc.x + size, 0);
			selectVec.addElement(pt);
			selectDim.width = 1;
			selectDim.height = 1;

			handlePaste();

			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			deselectVectorCells();

			for (int i = 0; i < backupSVec.size(); i++){				
				int ix = ((Point)backupSVec.elementAt(i)).x + size;
				int iy = ((Point)backupSVec.elementAt(i)).y ;
				Point np = new Point(ix, iy);
				selectVec.addElement(np);
				int r = ((Point)selectVec.elementAt(i)).y;
				int c = ((Point)selectVec.elementAt(i)).x;
				((Cell)cells.elementAt(c % columns + r * columns)).select();
			}

			ulc = findUpperLeftCorner();
			inputArea.setText( new String(((Cell)cells.elementAt(ulc.x % columns + ulc.y * columns)).getString()));

			for (int n = 0; n < backupCVec.size(); n++){
				copiedVec.addElement(backupCVec.elementAt(n));
				copiedVecStr.addElement(backupCVecStr.elementAt(n));				
			}

			selectDim.width = backupselectDim.width;
			selectDim.height = backupselectDim.height;
			copiedDim.width = backupcopiedDim.width;
			copiedDim.height = backupcopiedDim.height;
			
			fullUpdate = true;
			repaint();		
		}
	}

	public void handleRemoveRow(){
		int size = 0;
		if(selectDim.width != 0){
			size = selectVec.size()/selectDim.width;
		}else{
			return;
		}
		if((size > 0) && ((rows-size) > 0)){
			Point ulc = findUpperLeftCorner();
			
			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			copiedDim.width = 0;
			copiedDim.height = 0;
			deselectVectorCells();

			//index of row containing the upper left corner of the selection
			int indexOfULC = (ulc.y + 1) * columns;

			int where = indexOfULC - columns;
			for(int s=0; s < size; s++){
				for(int k=0; k<columns; k++){
					//System.out.println("where: " + where + " " + ((Cell)cells.elementAt(where)).getString() );
					cells.removeElementAt(where);	
				}
			}

			rows -= size;
			
			recalculate();
			inputArea.setText("");
			
			fullUpdate = true;
			repaint();
		}
	}
	

	public void handleRemoveColumn(){
		int size;
		if(selectDim.height != 0){
			size = selectVec.size()/selectDim.height;
		}else{
			return;
		}
		if((size > 0) && ((columns - size) > 0)){
			Point ulc = findUpperLeftCorner();

			copiedVec.removeAllElements();
			copiedVecStr.removeAllElements();
			copiedDim.width = 0;
			copiedDim.height = 0;
			deselectVectorCells();
			
			int tmpLineList[] = new int[columns-size];
			int ind=0;
			boolean removeFirstColumn = false;

			for(int n=0; n < m_dragLineList.length; n++){
				if(n < ulc.x){
					tmpLineList[ind] = m_dragLineList[ind];
					ind++;
				}else if(n >= (ulc.x + size)){
					//System.out.println(n + " width: " + getColumnWidth(n));
					
					if(removeFirstColumn == true){
						tmpLineList[ind] = getColumnWidth(n) + rowLabelWidth;
						removeFirstColumn = false;
					}else{
						//System.out.println("tmpLineList[ind-1]: " + tmpLineList[ind-1]);
						tmpLineList[ind] = tmpLineList[ind-1] + getColumnWidth(n);
					}
					ind++;
				}else{
					if(n == 0){
						removeFirstColumn = true;
					}
				}
			}

			int pos = ulc.x;
			
			for(int k=0; k<rows; k++){
				for(int s=0; s < size; s++){
					//System.out.println("pos: " + pos + " " + ((Cell)cells.elementAt(pos)).getString() );
					cells.removeElementAt(pos);
				}
				pos += columns - size;
			}

			columns -= size;
			
			m_dragLineList = new int[columns];
			System.arraycopy(tmpLineList, 0, m_dragLineList, 0, tmpLineList.length); 
			
			for(int in = 0; in < tmpLineList.length; in++){
				if(in == 0){
					setColumnWidth(in, tmpLineList[in] - rowLabelWidth);
				}else{
					setColumnWidth(in, tmpLineList[in] - tmpLineList[in-1]);
				}
			}

			recalculate();

			inputArea.setText("");

			fullUpdate = true;
			repaint();		
		}
	}

}
